# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import sys
import os
import grpc
from octopus_proto.agent_server_pb2_grpc import AgentServerServicer
from octopus_proto.agent_server_pb2_grpc import add_AgentServerServicer_to_server
from octopus_proto import agent_server_pb2
from octopus_proto import common_pb2
from octopus_proto import kernel_server_pb2
from dotenv import dotenv_values
from typing import AsyncIterable, Any, Dict, List, Optional, Sequence, Union, Type
from tempfile import gettempdir
from grpc.aio import ServicerContext, server
from octopus_kernel.sdk.kernel_sdk import KernelSDK
from .gpt_async_callback import AgentAsyncHandler
from .agent_llm import LLMManager
from .langchain_agent_builder import build_mock_agent, build_openai_agent

config = dotenv_values(".env")
LOG_LEVEL = logging.DEBUG
logging.basicConfig(
    level=LOG_LEVEL,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    handlers=[logging.StreamHandler(sys.stdout)],
)
logger = logging.getLogger(__name__)


class AgentRpcServer(AgentServerServicer):

    def __init__(self):
        self.agents = {}
        self.max_file_size = int(config["max_file_size"])
        self.verbose = config["verbose"]
        self.llm_manager = LLMManager(config)
        self.llm = self.llm_manager.get_llm()

    async def add_kernel(
        self, request: agent_server_pb2.AddKernelRequest, context: ServicerContext
    ) -> agent_server_pb2.AddKernelResponse:
        """Create a token, only the admin can call this method"""
        metadata = dict(context.invocation_metadata())
        if "api_key" not in metadata or metadata["api_key"] != config["admin_key"]:
            await context.abort(10, "You are not the admin")
        if request.key in self.agents and self.agents[request.key]:
            return agent_server_pb2.AddKernelResponse(code=0, msg="ok")
        # init the sdk
        sdk = KernelSDK(request.endpoint, request.key)
        sdk.connect()
        if config["llm_key"] == "azure_openai":
            logger.info("create a azure openai agent for kernel")
            agent = build_openai_agent(
                self.llm,
                sdk,
                config.get("max_iterations", 5),
                self.verbose,
            )
            # TODO a data dir per user
            self.agents[request.key] = {
                "sdk": sdk,
                "agent": agent,
            }
        elif config["llm_key"] == "mock":
            logger.info("create a mock agent for kernel")
            agent = build_mock_agent(self.llm)
            self.agents[request.key] = {
                "sdk": sdk,
                "agent": agent,
            }
        return agent_server_pb2.AddKernelResponse(code=0, msg="ok")

    async def send_task(
        self, request: agent_server_pb2.SendTaskRequest, context: ServicerContext
    ) -> AsyncIterable[agent_server_pb2.TaskRespond]:
        logger.debug("receive the task %s ", request.task)
        metadata = dict(context.invocation_metadata())
        if (
            "api_key" not in metadata
            or metadata["api_key"] not in self.agents
            or not self.agents[metadata["api_key"]]
        ):
            logger.debug("invalid api key")
            await context.abort(10, "invalid api key")
        agent = self.agents[metadata["api_key"]]["agent"]
        queue = asyncio.Queue()
        handler = AgentAsyncHandler(queue)

        async def worker(task, agent, handler):
            try:
                return await agent.arun(task, callbacks=[handler])
            except Exception as ex:
                logger.error("fail to run agent for %s", ex)
                result = str(ex)
                return result

        logger.debug("create the agent task")
        try:
            task = asyncio.create_task(worker(request.task, agent, handler))
            token_usage = 0
            while True:
                try:
                    logger.debug("start wait the queue message")
                    # TODO add timeout
                    respond = await queue.get()
                    if not respond:
                        logger.debug("exit the queue")
                        break
                    logger.debug(f"respond {respond}")
                    queue.task_done()
                    yield respond
                except Exception as ex:
                    logger.error(f"fail to get respond for {ex}")
                    break
            await task
            respond = agent_server_pb2.TaskRespond(
                token_usage=handler.token_usage,
                model_name=handler.model_name,
                iteration=handler.iteration,
                final_respond=agent_server_pb2.FinalRespond(answer=task.result()),
            )
            logger.debug(f"respond {respond}")
            yield respond
        except Exception as ex:
            respond = agent_server_pb2.TaskRespond(
                token_usage=0,
                model_name="",
                iteration=1,
                final_respond=agent_server_pb2.FinalRespond(answer=str(ex)),
            )
            yield respond

    async def download(
        self, request: common_pb2.DownloadRequest, context: ServicerContext
    ) -> AsyncIterable[common_pb2.FileChunk]:
        """
        download file
        """
        metadata = dict(context.invocation_metadata())
        if (
            "api_key" not in metadata
            or metadata["api_key"] not in self.agents
            or not self.agents[metadata["api_key"]]
        ):
            await context.abort(10, "invalid api key")
        sdk = self.agents[metadata["api_key"]]["sdk"]
        async for chunk in sdk.download_file(request.filename):
            yield chunk

    async def upload(
        self,
        request: AsyncIterable[common_pb2.FileChunk],
        context: ServicerContext,
    ) -> common_pb2.FileUploaded:
        """
        upload file
        """
        metadata = dict(context.invocation_metadata())
        if (
            "api_key" not in metadata
            or metadata["api_key"] not in self.agents
            or not self.agents[metadata["api_key"]]
        ):
            await context.abort(10, "invalid arguments")
        sdk = self.agents[metadata["api_key"]]["sdk"]

        async def generate_chunk(proxy_request, context, limit):
            length = 0
            async for chunk in proxy_request:
                if length + len(chunk.buffer) > limit:
                    await context.abort(
                        grpc.StatusCode.INVALID_ARGUMENT.value[0],
                        "exceed the max file limit",
                    )
                length += len(chunk.buffer)
                yield chunk

        return await sdk.upload_binary(
            generate_chunk(request, context, self.max_file_size)
        )


async def serve() -> None:
    logger.info(
        "start agent rpc server with host %s and port %s",
        config["rpc_host"],
        config["rpc_port"],
    )
    serv = server()
    add_AgentServerServicer_to_server(AgentRpcServer(), serv)
    listen_addr = "%s:%s" % (config["rpc_host"], config["rpc_port"])
    serv.add_insecure_port(listen_addr)
    await serv.start()
    await serv.wait_for_termination()


def server_main():
    asyncio.run(serve())



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" """

import logging
from langchain.chat_models import AzureChatOpenAI
from langchain.llms.fake import FakeListLLM

logger = logging.getLogger(__name__)


class LLMManager:

    def __init__(self, config):
        """
        llm_key=xxxx
        # the other optional
        """
        self.config = config
        self.llms = {}
        if self.config["llm_key"] == "azure_openai":
            self._build_azure_openai()
            self.llm_key = "azure_openai"
        elif self.config["llm_key"] == "mock":
            self._build_mock_llm()
            self.llm_key = "mock"

    def get_llm(self):
        return self.llms[self.llm_key]

    def get_llm_by_key(self, llm_key):
        """
        get llm with a key, the supported keys are 'mock', 'openai', 'azure_openai', 'codellama'
        """
        return self.llms[llm_key]

    def _no_empty_value_required(self, keys):
        for key in keys:
            if not self.config.get(key, None):
                raise Exception(f"the value of required {key} is empty")

    def _build_azure_openai(self):
        """
        build azure openai client from config
        """
        self._no_empty_value_required([
            "openai_api_base",
            "openai_api_version",
            "openai_api_key",
            "openai_api_type",
            "openai_api_deployment",
        ])
        api_base = self.config["openai_api_base"]
        api_version = self.config["openai_api_version"]
        api_type = self.config["openai_api_type"]
        api_key = self.config["openai_api_key"]
        api_deployment = self.config["openai_api_deployment"]
        temperature = self.config.get("temperature", 0)
        verbose = self.config.get("verbose", False)
        llm = AzureChatOpenAI(
            openai_api_base=api_base,
            openai_api_version=api_version,
            openai_api_key=api_key,
            openai_api_type=api_type,
            deployment_name=api_deployment,
            temperature=temperature,
            verbose=verbose,
        )
        self.llms["azure_openai"] = llm

    def _build_mock_llm(self):
        """
        build a mock llm
        """
        # the response to "how to get metadata from python grpc request"
        # TODO config the response
        responses = [
            """Final Answer: To get metadata from a Python gRPC request context, you can access the `context.invocation_metadata()` method. This method returns a list of key-value pairs representing the metadata associated with the request.

Here's an example of how you can retrieve metadata from a gRPC request context:

```python
def my_grpc_method(request, context):
    # Get the metadata from the request context
    metadata = dict(context.invocation_metadata())

    # Access specific metadata values
    value = metadata.get('key')

    # Print the metadata
    print(metadata)
```

In this example, `context.invocation_metadata()` returns a list of tuples representing the metadata. By converting it to a dictionary using `dict()`, you can easily access specific metadata values using their keys.

Note that the `context` parameter in the example represents the gRPC request context object passed to the gRPC method."""
        ]
        llm = FakeListLLM(responses=responses)
        self.llms["mock"] = llm



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from langchain.tools import StructuredTool
from .tools import OctopusAPIMarkdownOutput
from typing import Any, Dict, List, Optional, Sequence, Union, Type
from pydantic import BaseModel, Field
from langchain.callbacks.manager import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)


class AskQuestionInput(BaseModel):
    question: str = Field(description="the question")


class AskQuestionTool(StructuredTool):
    name = "ask_question"
    description = """ask the question from the human"""
    args_schema: Type[BaseModel] = AskQuestionInput
    return_direct = True

    def _run(
        self,
        question: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
        **kwargs: Any,
    ) -> Any:
        return "Yes"

    async def _arun(
        self,
        question: str,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> str:
        return "Yes"


class ExecutePythonCodeInput(BaseModel):
    code: str = Field(description="the python code to be executed")
    explanation: str = Field(description="the explanation of the python code")
    saved_filenames: Optional[List[str]] = Field(
        description="the saved filename list", default=[]
    )


class ExecutePythonCodeTool(StructuredTool):
    name = "execute_python_code"
    description = """Execute arbitrary Python code Returns a Markdown format string including return code, result, stdout, stderr, error"""
    args_schema: Type[BaseModel] = ExecutePythonCodeInput
    octopus_api: Optional[OctopusAPIMarkdownOutput] = None

    def _run(
        self,
        code: str,
        explanation: str,
        saved_filenames: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        code = code
        result = self.octopus_api.run(code)
        return result

    async def _arun(
        self,
        code: str,
        explanation: str,
        saved_filenames: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> str:
        return await self.octopus_api.arun(code)



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" """
import json
import asyncio
from uuid import UUID
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional
from langchain.callbacks.base import AsyncCallbackHandler, BaseCallbackHandler
from octopus_proto.agent_server_pb2 import OnAgentAction, TaskRespond, OnAgentActionEnd
from langchain.schema.agent import AgentAction, AgentFinish
from langchain.schema import LLMResult

logger = logging.getLogger(__name__)


class AgentAsyncHandler(AsyncCallbackHandler):
    """the agent async callback handler"""

    def __init__(self, queue):
        self.token_usage = 0
        self.model_name = ""
        self.iteration = 0
        self.images = []
        self.saved_images = []
        self.queue = queue
        self.stack = []

    async def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        logger.debug(f"on_tool_end {output}")
        self.stack.append("on_tool_end")
        output_files = []
        if output.find("the display data is saved to file") >= 0:
            output_files.append(output.split("`")[1])
        respond = TaskRespond(
            token_usage=self.token_usage,
            on_agent_action_end=OnAgentActionEnd(
                output=output, output_files=output_files
            ),
        )
        await self.queue.put(respond)

    async def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        logger.debug(f"on agent action {action}")
        self.stack.append("on_agent_action")
        respond = TaskRespond(
            token_usage=self.token_usage,
            on_agent_action=OnAgentAction(
                input=json.dumps(action.tool_input), tool=action.tool
            ),
        )
        await self.queue.put(respond)

    async def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        # end of the queue
        logger.debug(f"on_agent_finish {finish}")
        self.stack.append("on_agent_finish")
        logger.debug("the stack %s" % ("\n".join(self.stack)))
        await self.queue.put(None)

    async def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        self.stack.append("on_llm_start")

    async def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Run when LLM ends running."""
        logger.debug(f"on_llm_end {response.llm_output}")
        try:
            self.stack.append("on_llm_end")
            self.token_usage += response.llm_output["token_usage"]["total_tokens"]
            self.model_name = response.llm_output["model_name"]
            self.iteration += 1
        except Exception as ex:
            pass



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" """

from typing import Any, Dict, List, Optional, Sequence, Union, Type
from langchain.tools import BaseTool
from pydantic import BaseModel, Field
from langchain.callbacks.manager import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)


class PrintFinalAnswerInput(BaseModel):
    answer: str = Field(description="the final answer")


class PrintFinalAnswerTool(BaseTool):
    name = "print_final_answer"
    description = """print the the final answer"""
    args_schema: Type[BaseModel] = PrintFinalAnswerInput
    return_direct = True

    def _run(
        self,
        answer: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
        **kwargs: Any,
    ) -> Any:
        return ""

    async def _arun(
        self, answer: str, run_manager: Optional[AsyncCallbackManagerForToolRun] = None
    ) -> str:
        return ""



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" """
import asyncio
import json
import uuid
import base64
import logging
from typing import Any, Dict, List, Optional
from octopus_kernel.sdk.kernel_sdk import KernelSDK
from langchain.utils import get_from_dict_or_env

logger = logging.getLogger(__name__)


class OctopusAPIBase:
    """Wrap the Octopus Kernel API"""

    def __init__(self, sdk):
        self.sdk = sdk

    def run(self, code: str, **kwargs: Any) -> str:
        """Run python code

        Arguments:
        code  -- the python code to be executed
        """
        return asyncio.run(self.arun(code, **kwargs))

    async def arun(self, code: str, **kwargs: Any) -> str:
        """Run python code in async

        Arguments:
        code  -- the python code to be executed
        """
        kernel_name = (
            kwargs["kernel_name"]
            if "kernel_name" in kwargs and kwargs["kernel_name"]
            else "python3"
        )
        is_alive = await self.sdk.is_alive()
        if not is_alive:
            await self.sdk.start(kernel_name=kernel_name)
        messages = []
        response = await self.sdk.execute(code, kernel_name=kernel_name)
        # render the response to markdown
        output = self.render(response)
        return output

    def render(self, response, data_dir):
        raise NotImplementedError


class OctopusAPIMarkdownOutput(OctopusAPIBase):
    """Wrap the octopus kernel api with markdown output format

    You create the api key from octopus.dbpunk.com
    """

    def render(self, response):
        """
        reader the response to a markdown

        Args:
            response (object): The response object from the kernel client.

        Returns:
            str: The response data in Markdown format.
        """
        if response.traceback:
            return (
                "The code is `1` with the error\n```\n%s\n```"
                % json.loads(response.traceback)["data"]
            )

        output = ""
        if response.stdout:
            output = json.loads(response.stdout)["data"]
        if response.stderr:
            output += json.loads(response.stderr)["data"]
        if output:
            output = "The stdout/stderr \n```\n%s\n```" % output

        if response.result:
            result = json.loads(response.result)
            if result["msg_type"] == "execute_result":
                output = "The code is 0 and the result \n```\n%s\n``` \n %s" % (
                    result["data"]["text/plain"],
                    output,
                )
            elif result["msg_type"] == "display_data":
                if "image/png" in result["data"]:
                    filename = result["data"]["image/png"]
                    output = (
                        f"The code is 0 and the display data is saved to file `{filename}` \n%s"
                        % output
                    )
                elif "image/gif" in result["data"]:
                    filename = result["data"]["image/gif"]
                    output = (
                        f"The code is 0 and the display data is saved to file `{filename}` \n%s"
                        % output
                    )
                else:
                    keys = ",".join(result["data"].keys())
                    raise Exception(
                        f"unsupported display data type {keys} for the result"
                    )
            else:
                raise Exception(
                    f"unsupported messsage type {result['msg_type']} for the result"
                )
        return output


class OctopusAPIJsonOutput(OctopusAPIBase):
    """Wrap the octopus kernel api with json output format

    You create the api key from octopus.dbpunk.com
    """

    def render(self, response):
        """
        reader the response to a json

        Args:
            response (object): The response object from the kernel client.

        Returns:
            dict: The response data in JSON format.
        """
        output = {
            "code": 0,
            "result": None,
            "stdout": None,
            "stderr": None,
            "error": None,
        }
        if response.stdout:
            output["stdout"] = json.loads(response.stdout)["data"]
        if response.stderr:
            output["stderr"] = json.loads(response.stderr)["data"]
        if response.traceback:
            output["error"] = json.loads(response.traceback)["data"]
            output["code"] = 1
        if response.result:
            result = json.loads(response.result)
            if result["msg_type"] == "execute_result":
                output["result"] = result["data"]["text/plain"]
            elif result["msg_type"] == "display_data":
                if "image/png" in result["data"]:
                    filename = result["data"]["image/png"]
                    output[
                        "result"
                    ] = f"the image/png format data has been saved to file `{filename}`"
                elif "image/gif" in result["data"]:
                    filename = result["data"]["image/gif"]
                    output[
                        "result"
                    ] = f"the image/gif format data has been saved to file `{filename}`"
                else:
                    keys = ",".join(result["data"].keys())
                    raise Exception(
                        f"unsupported display data type {keys} for the result"
                    )
            else:
                raise Exception(
                    f"unsupported messsage type {result['msg_type']} for the result"
                )
        return output



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" """
import logging
import grpc
from grpc import aio
from octopus_proto import agent_server_pb2, common_pb2
from octopus_proto.agent_server_pb2_grpc import AgentServerStub
import aiofiles
from typing import AsyncIterable

logger = logging.getLogger(__name__)


class AgentSyncSDK:

    def __init__(self, endpoint, api_key):
        self.endpoint = endpoint
        self.stub = None
        self.metadata = aio.Metadata(
            ("api_key", api_key),
        )
        self.channel = None

    def connect(self):
        if self.channel:
            return
        self.channel = grpc.insecure_channel(self.endpoint)
        self.stub = AgentServerStub(self.channel)

    def add_kernel(self, key, endpoint):
        """
        add kernel instance to the agent and only admin can call this method
        """
        request = agent_server_pb2.AddKernelRequest(endpoint=endpoint, key=key)
        response = self.stub.add_kernel(request, metadata=self.metadata)
        return response

    def download_file(self, filename, parent_path):
        request = common_pb2.DownloadRequest(filename=filename)
        fullpath = "%s/%s" % (parent_path, filename)
        with open(fullpath, "wb+") as fd:
            for chunk in self.stub.download(request, metadata=self.metadata):
                fd.write(chunk.buffer)

    def upload_file(self, filepath, filename):
        """
        upload file to agent
        """

        # TODO limit the file size
        def generate_trunk(filepath, filename) -> common_pb2.FileChunk:
            try:
                with open(filepath, "rb") as fp:
                    while True:
                        chunk = fp.read(1024 * 128)
                        if not chunk:
                            break
                        yield common_pb2.FileChunk(buffer=chunk, filename=filename)
            except Exception as ex:
                logger.error("fail to read file %s" % ex)

        return self.stub.upload(
            generate_trunk(filepath, filename), metadata=self.metadata
        )

    def prompt(self, prompt, files=[]):
        """
        ask the ai with prompt and  uploaded files
        """
        request = agent_server_pb2.SendTaskRequest(task=prompt, input_files=files)
        for respond in self.stub.send_task(request, metadata=self.metadata):
            yield respond


class AgentSDK:

    def __init__(self, endpoint, api_key):
        self.endpoint = endpoint
        self.stub = None
        self.metadata = aio.Metadata(
            ("api_key", api_key),
        )
        self.channel = None

    def connect(self):
        """
        Connect the agent service
        """
        if self.channel:
            return
        channel = aio.insecure_channel(self.endpoint)
        self.channel = channel
        self.stub = AgentServerStub(channel)

    async def add_kernel(self, key, endpoint):
        """
        add kernel instance to the agent and only admin can call this method
        """
        request = agent_server_pb2.AddKernelRequest(endpoint=endpoint, key=key)
        response = await self.stub.add_kernel(request, metadata=self.metadata)
        return response

    async def prompt(self, prompt, files=[]):
        """
        ask the ai with prompt and  uploaded files
        """
        request = agent_server_pb2.SendTaskRequest(task=prompt, input_files=files)
        async for respond in self.stub.send_task(request, metadata=self.metadata):
            yield respond

    async def download_file(self, filename, parent_path):
        request = common_pb2.DownloadRequest(filename=filename)
        fullpath = "%s/%s" % (parent_path, filename)
        async with aiofiles.open(fullpath, "wb+") as afd:
            async for chunk in self.stub.download(request, metadata=self.metadata):
                await afd.write(chunk.buffer)

    async def upload_binary(self, chunks: AsyncIterable[common_pb2.FileChunk]):
        try:
            return await self.stub.upload(chunks, metadata=self.metadata)
        except Exception as ex:
            logger.error("upload file ex %s", ex)

    async def upload_file(self, filepath, filename):
        """
        upload file to agent
        """

        # TODO limit the file size
        async def generate_trunk(
            filepath, filename
        ) -> AsyncIterable[common_pb2.FileChunk]:
            try:
                async with aiofiles.open(filepath, "rb") as afp:
                    while True:
                        chunk = await afp.read(1024 * 128)
                        if not chunk:
                            break
                        yield common_pb2.FileChunk(buffer=chunk, filename=filename)
            except Exception as ex:
                logger.error("fail to read file %s", ex)

        await self.upload_binary(generate_trunk(filepath, filename))

    def close(self):
        if self.channel:
            self.channel.close()
            self.channel = None



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" """
from langchain.agents import initialize_agent
from langchain.schema.messages import SystemMessage
from .tools import OctopusAPIMarkdownOutput
from .gpt_tools import ExecutePythonCodeTool
from .mock_tools import PrintFinalAnswerTool as MockPrintFinalAnswerTool
from .prompt import OCTOPUS_FUNCTION_SYSTEM
from .gpt_async_callback import AgentAsyncHandler
from langchain.agents import AgentType


def build_openai_agent(llm, sdk, max_iterations, verbose):
    """build openai function call agent"""
    # TODO a data dir per user
    api = OctopusAPIMarkdownOutput(sdk)
    # init the agent
    tools = [
        ExecutePythonCodeTool(octopus_api=api),
    ]
    prefix = (
        """%sBegin!
Question: {input}
{agent_scratchpad}"""
        % OCTOPUS_FUNCTION_SYSTEM
    )
    system_message = SystemMessage(content=prefix)
    agent = initialize_agent(
        tools,
        llm,
        agent=AgentType.OPENAI_FUNCTIONS,
        agent_kwargs={"system_message": system_message},
        verbose=verbose,
        max_iterations=max_iterations,
        handle_parsing_errors=True,
    )
    return agent


def build_mock_agent(llm):
    tools = [
        MockPrintFinalAnswerTool(),
    ]
    agent = initialize_agent(tools, llm, agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION)
    return agent



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" """
from setuptools import setup, find_packages

setup(
    name="octopus_agent",
    version="0.1.4",
    description="Open source code interpreter agent for LLM",
    author="imotai",
    author_email="wangtaize@dbpunk.com",
    url="https://github.com/dbpunk-labs/octopus",
    packages=[
        "octopus_agent",
    ],
    package_dir={
        "octopus_agent": "src/octopus_agent",
    },
    install_requires=[
        "octopus_proto>=0.1.3",
        "octopus_kernel",
        "langchain>=0.0.227",
        "grpcio-tools>=1.57.0",
        "grpc-google-iam-v1>=0.12.6",
        "aiofiles",
        "pydantic",
        "python-dotenv",
        "openai",
        "aiosqlite",
    ],
    entry_points={
        "console_scripts": [
            "octopus_agent_rpc_server = octopus_agent.agent_server:server_main",
            "octopus_agent_setup = octopus_agent.agent_setup:setup",
        ]
    },
)



# vim:fenc=utf-8
#
# Copyright (C) 2023 dbpunk.com Author imotai <codego.me@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


OCTOPUS_FUNCTION_SYSTEM = """Firstly,You are the Programming Copilot called **Octopus**, a large language model designed to complete any goal by **executing code**

Secondly, Being an expert in programming, you must follow the rules
* To complete the goal, You must write a plan and execute it step by step, the followings are examples
    * The data visualization plan involves previewing, cleaning, and processing the data to generate the chart.
* Every step must include the explanation and the code block
* You must try to correct your code when you get errors from the output
* You are only allowed to ask yes or no questions from the human

Thirdly, the programming environment used to execute code has the following capabilities
* Internet connection: This allows the programming environment to access online resources, such as documentation, libraries, and code repositories.
* IPython kernel: This allows the programming environment to execute Python code
    * Filesystem: This allows the programming environment to open, write, and delete files in the workspace directory.
    * Lots of installed Python libraries:These includes the following popular libraries:
        * pandas:data analysis and manipulation tool
        * matplotlib:a comprehensive library for creating static, animated, and interactive visualizations in Python
        * yfinance:download market data from Yahoo!
        * imageio:library for reading and writing a wide range of image, video, scientific, and volumetric data formats.
        * pillow:Python Imaging Library
        * beautifulsoup4: a library that makes it easy to scrape information from web pages
        * requests: a simple, yet elegant, HTTP library
        * wikipedia: a Python library that makes it easy to access and parse data from Wikipedia
"""
